"""Random Vector Functional Link Network regression."""

# Authors: Thierry Moudiki
#
# License: BSD 3

import numpy as np
from .base import Base
import sklearn.metrics as skm
from ..utils import matrixops as mo
from ..utils import misc as mx
from ..utils import lmfuncs as lmf
from sklearn.base import RegressorMixin


class BaseRegressor(Base, RegressorMixin):
    """Base model with direct link and nonlinear activation.
        
       Parameters
       ----------
       n_hidden_features: int
           number of nodes in the hidden layer
       activation_name: str
           activation function: 'relu', 'tanh', 'sigmoid', 'prelu' or 'elu'
       a: float
           hyperparameter for 'prelu' or 'elu' activation function
       nodes_sim: str
           type of simulation for the nodes: 'sobol', 'hammersley', 'halton', 
           'uniform'
       bias: boolean
           indicates if the hidden layer contains a bias term (True) or 
           not (False)
       dropout: float
           regularization parameter; (random) percentage of nodes dropped out 
           of the training
       direct_link: boolean
           indicates if the original predictors are included (True) in model's 
           fitting or not (False)
       n_clusters: int
           number of clusters for type_clust='kmeans' or type_clust='gmm' 
           clustering (could be 0: no clustering)
       type_clust: str
           type of clustering method: currently k-means ('kmeans') or Gaussian 
           Mixture Model ('gmm')
       type_scaling: a tuple of 3 strings
           scaling methods for inputs, hidden layer, and clustering respectively
           (and when relevant). 
           Currently available: standardization ('std') or MinMax scaling ('minmax')
       col_sample: float
           percentage of covariates randomly chosen for training
       row_sample: float
           percentage of rows chosen for training, by stratified bootstrapping    
       seed: int 
           reproducibility seed for nodes_sim=='uniform', clustering and dropout
    """

    # construct the object -----

    def __init__(
        self,
        n_hidden_features=5,
        activation_name="relu",
        a=0.01,
        nodes_sim="sobol",
        bias=True,
        dropout=0,
        direct_link=True,
        n_clusters=2,
        type_clust="kmeans",
        type_scaling=("std", "std", "std"),
        col_sample=1,
        row_sample=1,
        seed=123,
    ):

        super().__init__(
            n_hidden_features=n_hidden_features,
            activation_name=activation_name,
            a=a,
            nodes_sim=nodes_sim,
            bias=bias,
            dropout=dropout,
            direct_link=direct_link,
            n_clusters=n_clusters,
            type_clust=type_clust,
            type_scaling=type_scaling,
            col_sample=col_sample,
            row_sample=row_sample,
            seed=seed,
        )

    def fit(self, X, y, **kwargs):
        """Fit RVFL to training data (X, y).
        
        Parameters
        ----------
        X: {array-like}, shape = [n_samples, n_features]
            Training vectors, where n_samples is the number 
            of samples and n_features is the number of features.
        
        y: array-like, shape = [n_samples]
               Target values.
    
        **kwargs: additional parameters to be passed to 
                  self.cook_training_set
               
        Returns
        -------
        self: object
        """
        centered_y, scaled_Z = self.cook_training_set(
            y=y, X=X, **kwargs
        )

        fit_obj = lmf.beta_Sigma_hat(
            X=scaled_Z, y=centered_y
        )

        self.beta = fit_obj["beta_hat"]

        self.GCV = fit_obj["GCV"]

        return self

    def predict(self, X, **kwargs):
        """Predict test data X.
        
        Parameters
        ----------
        X: {array-like}, shape = [n_samples, n_features]
            Training vectors, where n_samples is the number 
            of samples and n_features is the number of features.
        
        **kwargs: additional parameters to be passed to 
                  self.cook_test_set
               
        Returns
        -------
        model predictions: {array-like}
        """

        if len(X.shape) == 1:

            n_features = X.shape[0]
            new_X = mo.rbind(
                X.reshape(1, n_features),
                np.ones(n_features).reshape(1, n_features),
            )

            return (
                self.y_mean
                + np.dot(
                    self.cook_test_set(new_X, **kwargs),
                    self.beta,
                )
            )[0]

        else:

            return self.y_mean + np.dot(
                self.cook_test_set(X, **kwargs), self.beta
            )

    def score(self, X, y, scoring=None, **kwargs):
        """ Score the model on test set covariates X and response y. """

        preds = self.predict(X)

        if (
            type(preds) == tuple
        ):  # if there are std. devs in the predictions
            preds = preds[0]

        if scoring is None:
            scoring = "neg_mean_squared_error"

        # check inputs
        assert scoring in (
            "explained_variance",
            "neg_mean_absolute_error",
            "neg_mean_squared_error",
            "neg_mean_squared_log_error",
            "neg_median_absolute_error",
            "r2",
        ), "'scoring' should be in ('explained_variance', 'neg_mean_absolute_error', \
                           'neg_mean_squared_error', 'neg_mean_squared_log_error', \
                           'neg_median_absolute_error', 'r2')"

        scoring_options = {
            "explained_variance": skm.explained_variance_score,
            "neg_mean_absolute_error": skm.median_absolute_error,
            "neg_mean_squared_error": skm.mean_squared_error,
            "neg_mean_squared_log_error": skm.mean_squared_log_error,
            "neg_median_absolute_error": skm.median_absolute_error,
            "r2": skm.r2_score,
        }

        return scoring_options[scoring](y, preds, **kwargs)



"""Custom model"""

# Authors: Thierry Moudiki
#
# License: BSD 3

import numpy as np
import sklearn.metrics as skm2
from .custom import Custom
from ..utils import matrixops as mo
from sklearn.base import ClassifierMixin


class CustomClassifier(Custom, ClassifierMixin):
    """Custom Classification model class derived from class Custom
    
       Parameters
       ----------
       obj: object
           any object containing a method fit (obj.fit()) and a method predict 
           (obj.predict())
       n_hidden_features: int
           number of nodes in the hidden layer
       activation_name: str
           activation function: 'relu', 'tanh', 'sigmoid', 'prelu' or 'elu'
       a: float
           hyperparameter for 'prelu' or 'elu' activation function
       nodes_sim: str
           type of simulation for the nodes: 'sobol', 'hammersley', 'halton', 
           'uniform'
       bias: boolean
           indicates if the hidden layer contains a bias term (True) or not 
           (False)
       dropout: float
           regularization parameter; (random) percentage of nodes dropped out 
           of the training
       direct_link: boolean
           indicates if the original predictors are included (True) in model's 
           fitting or not (False)
       n_clusters: int
           number of clusters for 'kmeans' or 'gmm' clustering (could be 0: 
               no clustering)
       type_clust: str
           type of clustering method: currently k-means ('kmeans') or Gaussian 
           Mixture Model ('gmm')
       type_scaling: a tuple of 3 strings
           scaling methods for inputs, hidden layer, and clustering respectively
           (and when relevant). 
           Currently available: standardization ('std') or MinMax scaling ('minmax')
       col_sample: float
           percentage of covariates randomly chosen for training   
       row_sample: float
           percentage of rows chosen for training, by stratified bootstrapping    
       seed: int 
           reproducibility seed for nodes_sim=='uniform'
    """

    # construct the object -----

    def __init__(
        self,
        obj,
        n_hidden_features=5,
        activation_name="relu",
        a=0.01,
        nodes_sim="sobol",
        bias=True,
        dropout=0,
        direct_link=True,
        n_clusters=2,
        type_clust="kmeans",
        type_scaling=("std", "std", "std"),
        col_sample=1,
        row_sample=1,
        seed=123,
    ):

        super().__init__(
            obj=obj,
            n_hidden_features=n_hidden_features,
            activation_name=activation_name,
            a=a,
            nodes_sim=nodes_sim,
            bias=bias,
            dropout=dropout,
            direct_link=direct_link,
            n_clusters=n_clusters,
            type_clust=type_clust,
            type_scaling=type_scaling,
            col_sample=col_sample,
            row_sample=row_sample,
            seed=seed,
        )

        self.type_fit = "classification"

    def fit(self, X, y, **kwargs):
        """Fit custom model to training data (X, y).
        
        Parameters
        ----------
        X: {array-like}, shape = [n_samples, n_features]
            Training vectors, where n_samples is the number 
            of samples and n_features is the number of features.
        
        y: array-like, shape = [n_samples]
               Target values.
    
        **kwargs: additional parameters to be passed to 
                  self.cook_training_set or self.obj.fit
               
        Returns
        -------
        self: object
        """

        output_y, scaled_Z = self.cook_training_set(
            y=y, X=X, **kwargs
        )

        self.obj.fit(scaled_Z, output_y, **kwargs)

        return self

    def predict(self, X, **kwargs):
        """Predict test data X.
        
        Parameters
        ----------
        X: {array-like}, shape = [n_samples, n_features]
            Training vectors, where n_samples is the number 
            of samples and n_features is the number of features.
        
        **kwargs: additional parameters to be passed to 
                  self.cook_test_set
               
        Returns
        -------
        model predictions: {array-like}
        """

        if len(X.shape) == 1:

            n_features = X.shape[0]
            new_X = mo.rbind(
                X.reshape(1, n_features),
                np.ones(n_features).reshape(1, n_features),
            )

            return (
                self.obj.predict(
                    self.cook_test_set(new_X, **kwargs),
                    **kwargs
                )
            )[0]

        else:

            return self.obj.predict(
                self.cook_test_set(X, **kwargs), **kwargs
            )

    def predict_proba(self, X, **kwargs):
        """Predict probabilities for test data X.
        
        Parameters
        ----------
        X: {array-like}, shape = [n_samples, n_features]
            Training vectors, where n_samples is the number 
            of samples and n_features is the number of features.
        
        **kwargs: additional parameters to be passed to 
                  self.cook_test_set
               
        Returns
        -------
        probability estimates for test data: {array-like}        
        """
        try:

            if len(X.shape) == 1:

                n_features = X.shape[0]
                new_X = mo.rbind(
                    X.reshape(1, n_features),
                    np.ones(n_features).reshape(
                        1, n_features
                    ),
                )

                return (
                    self.obj.predict_proba(
                        self.cook_test_set(new_X, **kwargs),
                        **kwargs
                    )
                )[0]

            else:

                return self.obj.predict_proba(
                    self.cook_test_set(X, **kwargs),
                    **kwargs
                )

        except:

            raise ValueError(
                "No method 'predict_proba' found in object 'obj'"
            )

    def score(self, X, y, scoring=None, **kwargs):
        """ Score the model on test set covariates X and response y. """

        preds = self.predict(X)

        if scoring is None:
            scoring = "accuracy"

        # check inputs
        assert scoring in (
            "accuracy",
            "average_precision",
            "brier_score_loss",
            "f1",
            "f1_micro",
            "f1_macro",
            "f1_weighted",
            "f1_samples",
            "neg_log_loss",
            "precision",
            "recall",
            "roc_auc",
        ), "'scoring' should be in ('accuracy', 'average_precision', \
                           'brier_score_loss', 'f1', 'f1_micro', \
                           'f1_macro', 'f1_weighted',  'f1_samples', \
                           'neg_log_loss', 'precision', 'recall', \
                           'roc_auc')"

        scoring_options = {
            "accuracy": skm2.accuracy_score,
            "average_precision": skm2.average_precision_score,
            "brier_score_loss": skm2.brier_score_loss,
            "f1": skm2.f1_score,
            "f1_micro": skm2.f1_score,
            "f1_macro": skm2.f1_score,
            "f1_weighted": skm2.f1_score,
            "f1_samples": skm2.f1_score,
            "neg_log_loss": skm2.log_loss,
            "precision": skm2.precision_score,
            "recall": skm2.recall_score,
            "roc_auc": skm2.roc_auc_score,
        }

        return scoring_options[scoring](y, preds, **kwargs)



"""Tools for MTS."""

# Authors: Thierry Moudiki <thierry.moudiki@gmail.com>
#
# License: BSD 3

import numpy as np

# create inputs for training from MTS (envisage other regressors in X)
# X in decreasing order (!)
# a = np.reshape(range(0, 24), (8, 3))
# create_train_inputs(a, 2)
def create_train_inputs(X, k):

    n_k = X.shape[0] - k

    z = [X[i : n_k + i, :] for i in range(1, (k + 1))]

    return (X[0:n_k, :], np.column_stack(z))


# reformat response in prediction loop
# a = np.reshape(range(0, 24), (8, 3))
# reformat_response(a, 2)
def reformat_response(X, k):

    return np.concatenate([X[i, :] for i in range(k)])


# create k lags for series x
# def create_lags(x, k):
#
#    n = len(x)
#    n_k = n - k
#    n_col = k + 1
#    res = np.zeros((n_k, n_col))
#
#    for i in range(n_k):
#        for j in range(n_col):
#            res[i, j] = x[i + j]
#
#    return res


# create inputs for training from MTS (envisage other regressors in X)
# X in decreasing order (!)
# def create_train_inputs(X, k):
#
#    n, p = X.shape
#    n_k = n - k
#
#    y = np.zeros((n_k, p))
#    xreg = np.zeros((n_k, k*p))
#
#    for j in range(p):
#
#        Y_x = create_lags(X[:,j], k)
#
#        y[:,j] = Y_x[:, 0]
#
#        for l in range(1, k):
#
#            xreg[:, j*k + (l-1)] = Y_x[:, l]
#
#        xreg[:, j*k + (k-1)] = Y_x[:, k]
#
#    return (y, xreg)


# reformat response in prediction loop
# def reformat_response(X, k):
#
#    n, p = X.shape
#    res = np.zeros((1, k*p))
#
#    for j in range(p):
#        for i in range(k):
#            res[0, j*k + i] = X[i, j]
#
#    return res
#



from .base.base import Base
from .base.baseRegressor import BaseRegressor
from .custom.customClassifier import CustomClassifier
from .custom.customRegressor import CustomRegressor
from .mts import MTS
from .rvfl.bayesianrvflRegressor import (
    BayesianRVFLRegressor,
)
from .rvfl.bayesianrvfl2Regressor import (
    BayesianRVFL2Regressor,
)


__all__ = [
    "Base",
    "BaseRegressor",
    "BayesianRVFLRegressor",
    "BayesianRVFL2Regressor",
    "CustomClassifier",
    "CustomRegressor",
    "MTS",
]



"""Base class."""

# Authors: Thierry Moudiki
#
# License: BSD 3

import numpy as np
from functools import partial
from sklearn.base import BaseEstimator
from sklearn.cluster import KMeans
from sklearn.mixture import GaussianMixture
from sklearn.preprocessing import (
    StandardScaler,
    MinMaxScaler,
)
from ..utils import activations as ac
from ..utils import matrixops as mo
from ..utils import misc as mx
from ..simulation import nodesimulation as ns
from ..simulation import rowsubsampling as rs


class Base(BaseEstimator):
    """Base model with direct link and nonlinear activation.
        
       Parameters
       ----------
       n_hidden_features: int
           number of nodes in the hidden layer
       activation_name: str
           activation function: 'relu', 'tanh', 'sigmoid', 'prelu' or 'elu'
       a: float
           hyperparameter for 'prelu' or 'elu' activation function
       nodes_sim: str
           type of simulation for the nodes: 'sobol', 'hammersley', 'halton', 
           'uniform'
       bias: boolean
           indicates if the hidden layer contains a bias term (True) or 
           not (False)
       dropout: float
           regularization parameter; (random) percentage of nodes dropped out 
           of the training
       direct_link: boolean
           indicates if the original predictors are included (True) in model's 
           fitting or not (False)
       n_clusters: int
           number of clusters for type_clust='kmeans' or type_clust='gmm' 
           clustering (could be 0: no clustering)
       type_clust: str
           type of clustering method: currently k-means ('kmeans') or Gaussian 
           Mixture Model ('gmm')
       type_scaling: a tuple of 3 strings
           scaling methods for inputs, hidden layer, and clustering respectively
           (and when relevant). 
           Currently available: standardization ('std') or MinMax scaling ('minmax')
       col_sample: float
           percentage of covariates randomly chosen for training    
       row_sample: float
           percentage of rows chosen for training, by stratified bootstrapping    
       seed: int 
           reproducibility seed for nodes_sim=='uniform', clustering and dropout
    """

    # construct the object -----

    def __init__(
        self,
        n_hidden_features=5,
        activation_name="relu",
        a=0.01,
        nodes_sim="sobol",
        bias=True,
        dropout=0,
        direct_link=True,
        n_clusters=2,
        type_clust="kmeans",
        type_scaling=("std", "std", "std"),
        col_sample=1,
        row_sample=1,
        seed=123,
    ):

        # input checks -----

        assert activation_name in (
            "relu",
            "tanh",
            "sigmoid",
            "prelu",
            "elu",
        ), "'activation_name' must be in ('relu', 'tanh', 'sigmoid','prelu', 'elu')"

        assert nodes_sim in (
            "sobol",
            "hammersley",
            "uniform",
            "halton",
        ), "'nodes_sim' must be in ('sobol', 'hammersley', 'uniform', 'halton')"

        assert type_clust in (
            "kmeans",
            "gmm",
        ), "'type_clust' must be in ('kmeans', 'gmm')"

        assert (len(type_scaling) == 3) & all(
            type_scaling[i] in ("minmax", "std")
            for i in range(len(type_scaling))
        ), "'type_scaling' must have length 3, and available scaling methods are 'minmax' scaling and standardization ('std')"

        assert (col_sample >= 0) & (
            col_sample <= 1
        ), "'col_sample' must be comprised between 0 and 1 (both included)"

        # activation function -----

        activation_options = {
            "relu": ac.relu,
            "tanh": np.tanh,
            "sigmoid": ac.sigmoid,
            "prelu": partial(ac.prelu, a=a),
            "elu": partial(ac.elu, a=a),
        }

        self.n_hidden_features = n_hidden_features
        self.activation_name = activation_name
        self.a = a
        self.activation_func = activation_options[
            activation_name
        ]
        self.nodes_sim = nodes_sim
        self.bias = bias
        self.seed = seed
        self.dropout = dropout
        self.direct_link = direct_link
        self.type_clust = type_clust
        self.type_scaling = type_scaling
        self.col_sample = col_sample
        self.row_sample = row_sample
        self.index_col = None
        self.index_row = True
        self.n_clusters = n_clusters
        self.clustering_obj = None
        self.clustering_scaler = None
        self.nn_scaler = None
        self.scaler = None
        self.encoder = None
        self.W = None
        self.X = None
        self.y = None
        self.y_mean = None
        self.beta = None

    # "preprocessing" methods to be inherited -----

    def encode_clusters(
        self, X=None, predict=False, **kwargs
    ):  #
        """ Create new covariates with kmeans or GMM clustering. 

        Parameters
        ----------
        X: {array-like}, shape = [n_samples, n_features]
            Training vectors, where n_samples is the number 
            of samples and n_features is the number of features.
        
        predict: boolean
            is False on training set and True on test set
        
        **kwargs: 
            additional parameters to be passed to the 
            clustering method  
            
        Returns
        -------
        clusters' matrix, one-hot encoded: {array-like}        
        """

        if X is None:
            X = self.X

        if predict == False:  # encode training set

            scaling_options = {
                "std": StandardScaler(
                    copy=True, with_mean=True, with_std=True
                ),
                "minmax": MinMaxScaler(),
            }

            # scale input data before clustering
            scaler = scaling_options[self.type_scaling[2]]

            scaler.fit(X)
            scaled_X = scaler.transform(X)
            self.clustering_scaler = scaler

            if self.type_clust == "kmeans":

                # do kmeans + one-hot encoding
                np.random.seed(self.seed)
                kmeans = KMeans(
                    n_clusters=self.n_clusters, **kwargs
                )
                kmeans.fit(scaled_X)
                X_kmeans = kmeans.predict(scaled_X)
                self.clustering_obj = kmeans

                return mo.one_hot_encode(
                    X_kmeans, self.n_clusters
                )

            if self.type_clust == "gmm":

                np.random.seed(self.seed)
                gmm = GaussianMixture(
                    n_components=self.n_clusters, **kwargs
                )
                gmm.fit(scaled_X)
                X_gmm = gmm.predict(scaled_X)
                self.clustering_obj = gmm

                return mo.one_hot_encode(
                    X_gmm, self.n_clusters
                )

        else:  # if predict == True, encode test set

            X_clustered = self.clustering_obj.predict(
                self.clustering_scaler.transform(X)
            )

            return mo.one_hot_encode(
                X_clustered, self.n_clusters
            )

    def create_layer(self, scaled_X, W=None):
        """ Create hidden layer. """

        n_features = scaled_X.shape[1]

        if (
            self.bias != True
        ):  # no bias term in the hidden layer

            if W is None:

                if self.nodes_sim == "sobol":
                    self.W = ns.generate_sobol(
                        n_dims=n_features,
                        n_points=self.n_hidden_features,
                    )

                if self.nodes_sim == "hammersley":
                    self.W = ns.generate_hammersley(
                        n_dims=n_features,
                        n_points=self.n_hidden_features,
                    )

                if self.nodes_sim == "uniform":
                    self.W = ns.generate_uniform(
                        n_dims=n_features,
                        n_points=self.n_hidden_features,
                        seed=self.seed,
                    )

                if self.nodes_sim == "halton":
                    self.W = ns.generate_halton(
                        n_dims=n_features,
                        n_points=self.n_hidden_features,
                    )

                assert (
                    scaled_X.shape[1] == self.W.shape[0]
                ), "check dimensions of covariates X and matrix W"

                return mo.dropout(
                    x=self.activation_func(
                        np.dot(scaled_X, self.W)
                    ),
                    drop_prob=self.dropout,
                    seed=self.seed,
                )

            else:  # W is not none

                assert (
                    scaled_X.shape[1] == W.shape[0]
                ), "check dimensions of covariates X and matrix W"

                # self.W = W
                return mo.dropout(
                    x=self.activation_func(
                        np.dot(scaled_X, W)
                    ),
                    drop_prob=self.dropout,
                    seed=self.seed,
                )

        else:  # with bias term in the hidden layer

            if W is None:

                n_features_1 = n_features + 1

                if self.nodes_sim == "sobol":
                    self.W = ns.generate_sobol(
                        n_dims=n_features_1,
                        n_points=self.n_hidden_features,
                    )

                if self.nodes_sim == "hammersley":
                    self.W = ns.generate_hammersley(
                        n_dims=n_features_1,
                        n_points=self.n_hidden_features,
                    )

                if self.nodes_sim == "uniform":
                    self.W = ns.generate_uniform(
                        n_dims=n_features_1,
                        n_points=self.n_hidden_features,
                        seed=self.seed,
                    )

                if self.nodes_sim == "halton":
                    self.W = ns.generate_halton(
                        n_dims=n_features_1,
                        n_points=self.n_hidden_features,
                    )

                return mo.dropout(
                    x=self.activation_func(
                        np.dot(
                            mo.cbind(
                                np.ones(scaled_X.shape[0]),
                                scaled_X,
                            ),
                            self.W,
                        )
                    ),
                    drop_prob=self.dropout,
                    seed=self.seed,
                )

            else:

                # self.W = W
                return mo.dropout(
                    x=self.activation_func(
                        np.dot(
                            mo.cbind(
                                np.ones(scaled_X.shape[0]),
                                scaled_X,
                            ),
                            W,
                        )
                    ),
                    drop_prob=self.dropout,
                    seed=self.seed,
                )

    def cook_training_set(
        self, y=None, X=None, W=None, **kwargs
    ):
        """ Create new data for training set, with hidden layer, center the response. """

        # either X and y are stored or not
        # assert ((y is None) & (X is None)) | ((y is not None) & (X is not None))
        scaling_options = {
            "std": StandardScaler(
                copy=True, with_mean=True, with_std=True
            ),
            "minmax": MinMaxScaler(),
        }

        if self.n_hidden_features > 0:  # has a hidden layer

            assert len(self.type_scaling) >= 2, ""

            nn_scaling_options = {
                "std": StandardScaler(
                    copy=True, with_mean=True, with_std=True
                ),
                "minmax": MinMaxScaler(),
            }

            nn_scaler = nn_scaling_options[
                self.type_scaling[1]
            ]

        scaler = scaling_options[self.type_scaling[0]]

        # center y
        if mx.is_factor(y) == False:  # regression

            if y is None:
                y_mean = self.y.mean()
                centered_y = self.y - y_mean
            else:
                y_mean = y.mean()
                self.y_mean = y_mean
                centered_y = y - y_mean

        if X is None:

            if self.col_sample == 1:

                input_X = self.X

            else:

                n_features = self.X.shape[1]
                new_n_features = int(
                    np.ceil(n_features * self.col_sample)
                )
                assert (
                    new_n_features >= 1
                ), "check class attribute 'col_sample' and the number of covariates provided for X"
                np.random.seed(self.seed)
                index_col = np.random.choice(
                    range(n_features),
                    size=new_n_features,
                    replace=False,
                )
                self.index_col = index_col
                input_X = self.X[:, self.index_col]

        else:  # X is not None

            if self.col_sample == 1:

                input_X = X

            else:

                n_features = X.shape[1]
                new_n_features = int(
                    np.ceil(n_features * self.col_sample)
                )
                assert (
                    new_n_features >= 1
                ), "check class attribute 'col_sample' and the number of covariates provided for X"
                np.random.seed(self.seed)
                index_col = np.random.choice(
                    range(n_features),
                    size=new_n_features,
                    replace=False,
                )
                self.index_col = index_col
                input_X = X[:, self.index_col]

        if (
            self.n_clusters <= 0
        ):  # data without any clustering: self.n_clusters is None -----

            if (
                self.n_hidden_features > 0
            ):  # with hidden layer

                nn_scaler.fit(input_X)
                scaled_X = nn_scaler.transform(input_X)
                self.nn_scaler = nn_scaler

                if W is None:
                    Phi_X = self.create_layer(scaled_X)
                else:
                    Phi_X = self.create_layer(scaled_X, W=W)

                if self.direct_link == True:
                    Z = mo.cbind(input_X, Phi_X)
                else:
                    Z = Phi_X

                scaler.fit(Z)
                self.scaler = scaler

            else:  # no hidden layer

                Z = input_X
                scaler.fit(Z)
                self.scaler = scaler

        else:  # data with clustering: self.n_clusters is not None -----

            augmented_X = mo.cbind(
                input_X,
                self.encode_clusters(input_X, **kwargs),
            )

            if (
                self.n_hidden_features > 0
            ):  # with hidden layer

                nn_scaler.fit(augmented_X)
                scaled_X = nn_scaler.transform(augmented_X)
                self.nn_scaler = nn_scaler

                if W is None:
                    Phi_X = self.create_layer(scaled_X)
                else:
                    Phi_X = self.create_layer(scaled_X, W=W)

                if self.direct_link == True:
                    Z = mo.cbind(augmented_X, Phi_X)
                else:
                    Z = Phi_X

                scaler.fit(Z)
                self.scaler = scaler

            else:  # no hidden layer

                Z = augmented_X
                scaler.fit(Z)
                self.scaler = scaler

        # Returning model inputs -----

        # y is subsampled
        if self.row_sample < 1:

            n, p = Z.shape

            if y is None:

                self.index_row = rs.subsample(
                    y=self.y,
                    row_sample=self.row_sample,
                    seed=self.seed,
                )

            else:

                self.index_row = rs.subsample(
                    y=y,
                    row_sample=self.row_sample,
                    seed=self.seed,
                )

            n_row_sample = len(self.index_row)

            if mx.is_factor(y) == False:  # regression

                return (
                    centered_y[self.index_row].reshape(
                        n_row_sample
                    ),
                    self.scaler.transform(
                        Z[self.index_row, :].reshape(
                            n_row_sample, p
                        )
                    ),
                )

            else:  # classification

                return (
                    y[self.index_row].reshape(n_row_sample),
                    self.scaler.transform(
                        Z[self.index_row, :].reshape(
                            n_row_sample, p
                        )
                    ),
                )

        else:  # y is not subsampled

            if mx.is_factor(y) == False:  # regression

                return (
                    centered_y,
                    self.scaler.transform(Z),
                )

            else:  # classification

                return (y, self.scaler.transform(Z))

    def cook_test_set(self, X, **kwargs):
        """ Transform data from test set, with hidden layer. """

        if (
            self.n_clusters <= 0
        ):  # data without clustering: self.n_clusters is None -----

            if (
                self.n_hidden_features > 0
            ):  # if hidden layer

                if self.col_sample == 1:

                    scaled_X = self.nn_scaler.transform(X)

                else:

                    scaled_X = self.nn_scaler.transform(
                        X[:, self.index_col]
                    )

                Phi_X = self.create_layer(scaled_X, self.W)

                if self.direct_link == True:
                    return self.scaler.transform(
                        mo.cbind(X, Phi_X)
                    )
                else:
                    return self.scaler.transform(Phi_X)

            else:  # if no hidden layer

                return self.scaler.transform(X)

        else:  # data with clustering: self.n_clusters is None -----

            if self.col_sample == 1:

                predicted_clusters = self.encode_clusters(
                    X=X, predict=True, **kwargs
                )
                augmented_X = mo.cbind(
                    X, predicted_clusters
                )

            else:

                predicted_clusters = self.encode_clusters(
                    X=X[:, self.index_col],
                    predict=True,
                    **kwargs
                )
                augmented_X = mo.cbind(
                    X[:, self.index_col], predicted_clusters
                )

            if (
                self.n_hidden_features > 0
            ):  # if hidden layer

                scaled_X = self.nn_scaler.transform(
                    augmented_X
                )
                Phi_X = self.create_layer(scaled_X, self.W)

                if self.direct_link == True:
                    return self.scaler.transform(
                        mo.cbind(augmented_X, Phi_X)
                    )
                else:
                    return self.scaler.transform(Phi_X)

            else:  # if no hidden layer

                return self.scaler.transform(augmented_X)



import numpy as np


# column bind
def cbind(x, y):

    # if len(x.shape) == 1 or len(y.shape) == 1:
    return np.column_stack((x, y))


# computes t(x)%*%y
def crossprod(x, y=None):
    # assert on dimensions
    if y is None:
        return np.dot(x.transpose(), x)
    else:
        return np.dot(x.transpose(), y)


# dropout
def dropout(x, drop_prob=0, seed=123):

    assert 0 <= drop_prob <= 1

    n, p = x.shape

    if drop_prob == 0:
        return x

    if drop_prob == 1:
        return np.zeros_like(x)

    np.random.seed(seed)
    dropped_indices = np.random.rand(n, p) > drop_prob

    return dropped_indices * x / (1 - drop_prob)


# one-hot encoding
def one_hot_encode(x_clusters, n_clusters):

    assert max(x_clusters) <= n_clusters

    n_obs = len(x_clusters)
    res = np.zeros((n_obs, n_clusters))

    for i in range(n_obs):
        res[i, x_clusters[i]] = 1

    return res


# row bind
def rbind(x, y):

    # if len(x.shape) == 1 or len(y.shape) == 1:
    return np.row_stack((x, y))


# else:
#    return np.concatenate((x, y),
#                      axis = 0)

# computes x%*%t(y)
def tcrossprod(x, y=None):
    # assert on dimensions
    if y is None:
        return np.dot(x, x.transpose())
    else:
        return np.dot(x, y.transpose())


# convert vector to numpy array
def to_np_array(X):
    return np.array(X.copy(), ndmin=2)


# scale matrix
# def scale_matrix(X, x_means=None, x_vars=None):
#
#    if ((x_means is None) & (x_vars is None)):
#        x_means = X.mean(axis = 0)
#        x_vars = X.var(axis = 0)
#        return ((X - x_means)/np.sqrt(x_vars),
#                x_means,
#                x_vars)
#
#    if ((x_means is not None) & (x_vars is None)):
#        return X - x_means
#
#    if ((x_means is None) & (x_vars is not None)):
#        return X/np.sqrt(x_vars)
#
#    if ((x_means is not None) & (x_vars is not None)):
#        return (X - x_means)/np.sqrt(x_vars)



import numpy as np


# merge two dictionaries
def merge_two_dicts(x, y):
    z = x.copy()
    z.update(y)
    return z


# check if the response contains only integers
def is_factor(y):

    n = len(y)

    cond0 = (
        sum(
            list(
                map(
                    lambda x: isinstance(y[x], int),
                    range(n),
                )
            )
        )
        == n
    )
    cond1 = (
        sum(
            list(
                map(
                    lambda x: isinstance(y[x], np.int),
                    range(n),
                )
            )
        )
        == n
    )
    cond2 = (
        sum(
            list(
                map(
                    lambda x: isinstance(y[x], np.int32),
                    range(n),
                )
            )
        )
        == n
    )
    cond3 = (
        sum(
            list(
                map(
                    lambda x: isinstance(y[x], np.int64),
                    range(n),
                )
            )
        )
        == n
    )

    return cond0 or cond1 or cond2 or cond3


# flatten list of lists
flatten = lambda l: [
    item for sublist in l for item in sublist
]



import numpy as np
from numpy import linalg as la

# import .matrixops as mo
from . import matrixops as mo


# computes beta_hat = (t(x)%*%x + lam*I)^{-1}%*%t(x)%*%y
def beta_hat(x, y, lam=None):
    # assert on dimensions
    if lam is None:
        return np.dot(
            inv_penalized_cov(x), mo.crossprod(x, y)
        )
    else:
        return np.dot(
            inv_penalized_cov(x, lam), mo.crossprod(x, y)
        )


# computes (t(x)%*%x + lam*I)^{-1}
def inv_penalized_cov(x, lam=None):
    # assert on dimensions
    if lam is None:
        return la.inv(mo.crossprod(x))
    else:
        return la.inv(
            mo.crossprod(x) + lam * np.eye(x.shape[1])
        )


# linear regression with no regularization
def beta_Sigma_hat(
    X=None,
    y=None,
    fit_intercept=False,
    X_star=None,  # for preds
    return_cov=True,  # confidence interval required for preds?
    beta_hat_=None,  # for prediction only (X_star is not None)
    Sigma_hat_=None,
):  # for prediction only (X_star is not None)

    if (X is not None) & (y is not None):  # fit

        if len(X.shape) == 1:
            X = X.reshape(-1, 1)

        if X_star is not None:
            if len(X_star.shape) == 1:
                X_star = X_star.reshape(-1, 1)

        n, p = X.shape

        if fit_intercept == True:
            X = mo.cbind(np.ones(n), X)
            if X_star is not None:
                X_star = mo.cbind(
                    np.ones(X_star.shape[0]), X_star
                )

        Cn = inv_penalized_cov(X)

        if return_cov == True:

            if X_star is None:

                beta_hat_ = np.dot(Cn, mo.crossprod(X, y))
                Sigma_hat_ = np.eye(X.shape[1]) - np.dot(
                    Cn, mo.crossprod(X)
                )
                temp = np.dot(
                    Cn, mo.tcrossprod(Sigma_hat_, X)
                )
                smoothing_matrix = np.dot(X, temp)
                y_hat = np.dot(smoothing_matrix, y)

                return {
                    "beta_hat": beta_hat_,
                    "Sigma_hat": Sigma_hat_,
                    "GCV": np.mean(
                        (
                            (y - y_hat)
                            / (
                                1
                                - np.trace(smoothing_matrix)
                                / n
                            )
                        )
                        ** 2
                    ),
                }

            else:

                if beta_hat_ is None:
                    beta_hat_ = np.dot(
                        Cn, mo.crossprod(X, y)
                    )

                if Sigma_hat_ is None:
                    Sigma_hat_ = np.eye(
                        X.shape[1]
                    ) - np.dot(Cn, mo.crossprod(X))

                temp = np.dot(
                    Cn, mo.tcrossprod(Sigma_hat_, X)
                )
                smoothing_matrix = np.dot(X, temp)
                y_hat = np.dot(smoothing_matrix, y)

                return {
                    "beta_hat": beta_hat_,
                    "Sigma_hat": Sigma_hat_,
                    "GCV": np.mean(
                        (
                            (y - y_hat)
                            / (
                                1
                                - np.trace(smoothing_matrix)
                                / n
                            )
                        )
                        ** 2
                    ),
                    "preds": np.dot(X_star, beta_hat_),
                    "preds_std": np.sqrt(
                        np.diag(
                            np.dot(
                                X_star,
                                mo.tcrossprod(
                                    Sigma_hat_, X_star
                                ),
                            )
                        )
                    ),
                }

        else:  # return_cov == False

            if X_star is None:

                beta_hat_ = np.dot(Cn, mo.crossprod(X, y))
                Sigma_hat_ = np.eye(X.shape[1]) - np.dot(
                    Cn, mo.crossprod(X)
                )
                temp = np.dot(
                    Cn, mo.tcrossprod(Sigma_hat_, X)
                )
                smoothing_matrix = np.dot(X, temp)
                y_hat = np.dot(smoothing_matrix, y)

                return {
                    "beta_hat": beta_hat_,
                    "GCV": np.mean(
                        (
                            (y - y_hat)
                            / (
                                1
                                - np.trace(smoothing_matrix)
                                / n
                            )
                        )
                        ** 2
                    ),
                }

            else:

                if beta_hat_ is None:
                    beta_hat_ = beta_hat(X, y)
                    Sigma_hat_ = np.eye(
                        X.shape[1]
                    ) - np.dot(Cn, mo.crossprod(X))
                    temp = np.dot(
                        Cn, mo.tcrossprod(Sigma_hat_, X)
                    )
                    smoothing_matrix = np.dot(X, temp)
                    y_hat = np.dot(smoothing_matrix, y)

                return {
                    "beta_hat": beta_hat_,
                    "GCV": np.mean(
                        (
                            (y - y_hat)
                            / (
                                1
                                - np.trace(smoothing_matrix)
                                / n
                            )
                        )
                        ** 2
                    ),
                    "preds": np.dot(X_star, beta_hat_),
                }

    else:  # X is None | y is None  # predict

        assert (beta_hat_ is not None) & (
            X_star is not None
        )

        if return_cov == True:

            assert Sigma_hat_ is not None

            return {
                "preds": np.dot(X_star, beta_hat_),
                "preds_std": np.sqrt(
                    np.diag(
                        np.dot(
                            X_star,
                            mo.tcrossprod(
                                Sigma_hat_, X_star
                            ),
                        )
                    )
                ),
            }

        else:

            return {"preds": np.dot(X_star, beta_hat_)}


# beta and Sigma in Bayesian Ridge Regression 1
# without intercept! without intercept! without intercept!
def beta_Sigma_hat_rvfl(
    X=None,
    y=None,
    s=0.1,
    sigma=0.05,
    fit_intercept=False,
    X_star=None,  # for preds
    return_cov=True,  # confidence interval required for preds?
    beta_hat_=None,  # for prediction only (X_star is not None)
    Sigma_hat_=None,
):  # for prediction only (X_star is not None)

    if (X is not None) & (y is not None):  # fit

        if len(X.shape) == 1:
            X = X.reshape(-1, 1)

        if X_star is not None:
            if len(X_star.shape) == 1:
                X_star = X_star.reshape(-1, 1)

        n, p = X.shape

        if fit_intercept == True:
            X = mo.cbind(np.ones(n), X)
            if X_star is not None:
                X_star = mo.cbind(
                    np.ones(X_star.shape[0]), X_star
                )

        s2 = s ** 2
        lambda_ = (sigma ** 2) / s2
        Cn = inv_penalized_cov(X, lam=lambda_)

        if return_cov == True:

            if X_star is None:

                beta_hat_ = np.dot(Cn, mo.crossprod(X, y))
                Sigma_hat_ = s2 * (
                    np.eye(X.shape[1])
                    - np.dot(Cn, mo.crossprod(X))
                )
                temp = np.dot(
                    Cn, mo.tcrossprod(Sigma_hat_, X)
                )
                smoothing_matrix = np.dot(X, temp)
                y_hat = np.dot(smoothing_matrix, y)

                return {
                    "beta_hat": beta_hat_,
                    "Sigma_hat": Sigma_hat_,
                    "GCV": np.mean(
                        (
                            (y - y_hat)
                            / (
                                1
                                - np.trace(smoothing_matrix)
                                / n
                            )
                        )
                        ** 2
                    ),
                }

            else:

                if beta_hat_ is None:
                    beta_hat_ = np.dot(
                        Cn, mo.crossprod(X, y)
                    )

                if Sigma_hat_ is None:
                    Sigma_hat_ = s2 * (
                        np.eye(X.shape[1])
                        - np.dot(Cn, mo.crossprod(X))
                    )

                temp = np.dot(
                    Cn, mo.tcrossprod(Sigma_hat_, X)
                )
                smoothing_matrix = np.dot(X, temp)
                y_hat = np.dot(smoothing_matrix, y)

                return {
                    "beta_hat": beta_hat_,
                    "Sigma_hat": Sigma_hat_,
                    "GCV": np.mean(
                        (
                            (y - y_hat)
                            / (
                                1
                                - np.trace(smoothing_matrix)
                                / n
                            )
                        )
                        ** 2
                    ),
                    "preds": np.dot(X_star, beta_hat_),
                    "preds_std": np.sqrt(
                        np.diag(
                            np.dot(
                                X_star,
                                mo.tcrossprod(
                                    Sigma_hat_, X_star
                                ),
                            )
                            + (sigma ** 2)
                            * np.eye(X_star.shape[0])
                        )
                    ),
                }

        else:  # return_cov == False

            if X_star is None:

                beta_hat_ = np.dot(Cn, mo.crossprod(X, y))
                Sigma_hat_ = s2 * (
                    np.eye(X.shape[1])
                    - np.dot(Cn, mo.crossprod(X))
                )
                temp = np.dot(
                    Cn, mo.tcrossprod(Sigma_hat_, X)
                )
                smoothing_matrix = np.dot(X, temp)
                y_hat = np.dot(smoothing_matrix, y)

                return {
                    "beta_hat": beta_hat_,
                    "GCV": np.mean(
                        (
                            (y - y_hat)
                            / (
                                1
                                - np.trace(smoothing_matrix)
                                / n
                            )
                        )
                        ** 2
                    ),
                }

            else:

                if beta_hat_ is None:
                    beta_hat_ = beta_hat(X, y, lam=lambda_)
                    Sigma_hat_ = s2 * (
                        np.eye(X.shape[1])
                        - np.dot(Cn, mo.crossprod(X))
                    )
                    temp = np.dot(
                        Cn, mo.tcrossprod(Sigma_hat_, X)
                    )
                    smoothing_matrix = np.dot(X, temp)
                    y_hat = np.dot(smoothing_matrix, y)

                return {
                    "beta_hat": beta_hat_,
                    "GCV": np.mean(
                        (
                            (y - y_hat)
                            / (
                                1
                                - np.trace(smoothing_matrix)
                                / n
                            )
                        )
                        ** 2
                    ),
                    "preds": np.dot(X_star, beta_hat_),
                }

    else:  # X is None | y is None  # predict

        assert (beta_hat_ is not None) & (
            X_star is not None
        )

        if return_cov == True:

            assert Sigma_hat_ is not None

            return {
                "preds": np.dot(X_star, beta_hat_),
                "preds_std": np.sqrt(
                    np.diag(
                        np.dot(
                            X_star,
                            mo.tcrossprod(
                                Sigma_hat_, X_star
                            ),
                        )
                        + (sigma ** 2)
                        * np.eye(X_star.shape[0])
                    )
                ),
            }

        else:

            return {"preds": np.dot(X_star, beta_hat_)}


# beta and Sigma in Bayesian Ridge Regression 2
# without intercept! without intercept! without intercept!
def beta_Sigma_hat_rvfl2(
    X=None,
    y=None,
    Sigma=None,
    sigma=0.05,
    fit_intercept=False,
    X_star=None,  # check when dim = 1 # check when dim = 1
    return_cov=True,
    beta_hat_=None,  # for prediction only (X_star is not None)
    Sigma_hat_=None,
):  # for prediction only (X_star is not None)

    if (X is not None) & (y is not None):

        if len(X.shape) == 1:
            X = X.reshape(-1, 1)

        n, p = X.shape

        if Sigma is None:
            if fit_intercept == True:
                Sigma = np.eye(p + 1)
            else:
                Sigma = np.eye(p)

        if X_star is not None:
            if len(X_star.shape) == 1:
                X_star = X_star.reshape(-1, 1)

        if fit_intercept == True:

            X = mo.cbind(np.ones(n), X)
            Cn = la.inv(
                np.dot(Sigma, mo.crossprod(X))
                + (sigma ** 2) * np.eye(p + 1)
            )

            if X_star is not None:
                X_star = mo.cbind(
                    np.ones(X_star.shape[0]), X_star
                )
        else:
            # rename to invCn
            Cn = la.inv(
                np.dot(Sigma, mo.crossprod(X))
                + (sigma ** 2) * np.eye(p)
            )

        temp = np.dot(Cn, mo.tcrossprod(Sigma, X))
        smoothing_matrix = np.dot(X, temp)
        y_hat = np.dot(smoothing_matrix, y)

        if return_cov == True:

            if X_star is None:

                return {
                    "beta_hat": np.dot(temp, y),
                    "Sigma_hat": Sigma
                    - np.dot(temp, np.dot(X, Sigma)),
                    "GCV": np.mean(
                        (
                            (y - y_hat)
                            / (
                                1
                                - np.trace(smoothing_matrix)
                                / n
                            )
                        )
                        ** 2
                    ),
                }

            else:

                if beta_hat_ is None:
                    beta_hat_ = np.dot(temp, y)

                if Sigma_hat_ is None:
                    Sigma_hat_ = Sigma - np.dot(
                        temp, np.dot(X, Sigma)
                    )

                return {
                    "beta_hat": beta_hat_,
                    "Sigma_hat": Sigma_hat_,
                    "GCV": np.mean(
                        (
                            (y - y_hat)
                            / (
                                1
                                - np.trace(smoothing_matrix)
                                / n
                            )
                        )
                        ** 2
                    ),
                    "preds": np.dot(X_star, beta_hat_),
                    "preds_std": np.sqrt(
                        np.diag(
                            np.dot(
                                X_star,
                                mo.tcrossprod(
                                    Sigma_hat_, X_star
                                ),
                            )
                            + (sigma ** 2)
                            * np.eye(X_star.shape[0])
                        )
                    ),
                }

        else:  # return_cov == False

            if X_star is None:

                return {
                    "beta_hat": np.dot(temp, y),
                    "GCV": np.mean(
                        (
                            (y - y_hat)
                            / (
                                1
                                - np.trace(smoothing_matrix)
                                / n
                            )
                        )
                        ** 2
                    ),
                }

            else:

                if beta_hat_ is None:
                    beta_hat_ = np.dot(temp, y)

                return {
                    "beta_hat": beta_hat_,
                    "GCV": np.mean(
                        (
                            (y - y_hat)
                            / (
                                1
                                - np.trace(smoothing_matrix)
                                / n
                            )
                        )
                        ** 2
                    ),
                    "preds": np.dot(X_star, beta_hat_),
                }

    else:  # (X is None) | (y is None) # predict

        assert (beta_hat_ is not None) & (
            X_star is not None
        )

        if return_cov == True:

            assert Sigma_hat_ is not None

            return {
                "preds": np.dot(X_star, beta_hat_),
                "preds_std": np.sqrt(
                    np.diag(
                        np.dot(
                            X_star,
                            mo.tcrossprod(
                                Sigma_hat_, X_star
                            ),
                        )
                        + (sigma ** 2)
                        * np.eye(X_star.shape[0])
                    )
                ),
            }

        else:

            return {"preds": np.dot(X_star, beta_hat_)}



"""MTS model"""

# Authors: Thierry Moudiki
#
# License: BSD 3

# ts objects with rpy2
# ts objects with rpy2
# ts objects with rpy2
# ts objects with rpy2
# ts objects with rpy2

import numpy as np
from ..base import Base
from scipy.stats import norm
from ..utils import matrixops as mo
from ..utils import timeseries as ts
import sklearn.metrics as skm2


class MTS(Base):
    """MTS model class derived from class Base
    
       Parameters
       ----------
       obj: object
           any object containing a method fit (obj.fit()) and a method predict 
           (obj.predict())
       n_hidden_features: int
           number of nodes in the hidden layer
       activation_name: str
           activation function: 'relu', 'tanh', 'sigmoid', 'prelu' or 'elu'
       a: float
           hyperparameter for 'prelu' or 'elu' activation function
       nodes_sim: str
           type of simulation for the nodes: 'sobol', 'hammersley', 'halton', 
           'uniform'
       bias: boolean
           indicates if the hidden layer contains a bias term (True) or not 
           (False)
       dropout: float
           regularization parameter; (random) percentage of nodes dropped out 
           of the training
       direct_link: boolean
           indicates if the original predictors are included (True) in model's 
           fitting or not (False)
       n_clusters: int
           number of clusters for 'kmeans' or 'gmm' clustering (could be 0: 
               no clustering)
       type_clust: str
           type of clustering method: currently k-means ('kmeans') or Gaussian 
           Mixture Model ('gmm')
       type_scaling: a tuple of 3 strings
           scaling methods for inputs, hidden layer, and clustering respectively
           (and when relevant). 
           Currently available: standardization ('std') or MinMax scaling ('minmax')
       col_sample: float
           percentage of covariates randomly chosen for training    
       seed: int 
           reproducibility seed for nodes_sim=='uniform'
       lags: int
           number of lags for the time series 
    """

    # construct the object -----

    def __init__(
        self,
        obj,
        n_hidden_features=5,
        activation_name="relu",
        a=0.01,
        nodes_sim="sobol",
        bias=True,
        dropout=0,
        direct_link=True,
        n_clusters=2,
        type_clust="kmeans",
        type_scaling=("std", "std", "std"),
        col_sample=1,
        seed=123,
        lags=1,
        return_std=False,
    ):

        assert (
            np.int(lags) == lags
        ), "parameter 'lags' should be an integer"

        super().__init__(
            n_hidden_features=n_hidden_features,
            activation_name=activation_name,
            a=a,
            nodes_sim=nodes_sim,
            bias=bias,
            dropout=dropout,
            direct_link=direct_link,
            n_clusters=n_clusters,
            type_clust=type_clust,
            type_scaling=type_scaling,
            col_sample=col_sample,
            seed=seed,
        )

        self.obj = obj
        self.n_series = None
        self.lags = lags
        self.fit_objs = {}
        self.y = (
            None
        )  # MTS responses (most recent observations first)
        self.X = None  # MTS lags
        self.xreg = None
        self.regressors_scaler = None
        self.y_means = []
        self.preds = None
        self.return_std = return_std
        self.preds_std = []
        self.row_sample = 1

    def fit(self, X, xreg=None, **kwargs):
        """Fit MTS model to training data (X, y).
        
        Parameters
        ----------
        X: {array-like}, shape = [n_samples, n_features]
            Training time series, where n_samples is the number 
            of samples and n_features is the number of features.
            X must be in increasing order (most recent observations last)
        xreg: {array-like}, shape = [n_samples, n_features_xreg]
              Additional regressors to be passed to obj
              xreg must be in increasing order (most recent observations last)
        **kwargs: additional parameters to be passed to 
                  self.cook_training_set
               
        Returns
        -------
        self: object
        """

        self.y = None

        self.X = None

        n, p = X.shape

        self.n_series = p

        self.y_means = np.zeros(p)

        mts_input = ts.create_train_inputs(
            X[::-1], self.lags
        )

        self.y = mts_input[0]

        self.X = mts_input[1]

        if xreg is not None:

            assert (
                xreg.shape[0] == n
            ), "'xreg' and 'X' must have the same number or observations"

            self.xreg = xreg

            xreg_input = ts.create_train_inputs(
                xreg[::-1], self.lags
            )

            dummy_y, scaled_Z = self.cook_training_set(
                y=np.repeat(1, n),
                X=mo.cbind(self.X, xreg_input[1]),
                **kwargs
            )

        else:  # xreg is None

            # avoids scaling X p times in the loop
            dummy_y, scaled_Z = self.cook_training_set(
                y=np.repeat(1, n), X=self.X, **kwargs
            )

        # loop on all the time series and adjust self.obj.fit
        for i in range(p):
            y_mean = np.mean(self.y[:, i])
            self.y_means[i] = y_mean
            self.fit_objs[i] = self.obj.fit(
                scaled_Z, self.y[:, i] - y_mean, **kwargs
            )

        self.y_mean = None

        return self

    def predict(
        self, h=5, level=95, new_xreg=None, **kwargs
    ):
        """Predict on horizon h.
        
        Parameters
        ----------
        h: {integer}
            Forecasting horizon
        
        level: {integer}
            Level of confidence (if obj has option 'return_std' and the 
            posterior is gaussian)
            
        new_xreg: {array-like}, shape = [n_samples = h, n_new_xreg]
            New values of additional (deterministic) regressors on horizon = h
            new_xreg must be in increasing order (most recent observations last)
        
        **kwargs: additional parameters to be passed to 
                  self.cook_test_set
               
        Returns
        -------
        model predictions for horizon = h: {array-like}
        """

        if self.xreg is not None:
            assert (
                new_xreg is not None
            ), "'new_xreg' must be provided"

        self.preds = None

        self.preds = self.y

        n_features = self.n_series * self.lags

        self.preds_std = np.zeros(h)

        if "return_std" in kwargs:

            self.return_std = kwargs["return_std"]

            qt = norm.ppf(1 - 0.5 * (1 - level / 100))

        if (
            new_xreg is not None
        ):  # Additional regressors provided

            try:
                n_obs_xreg, n_features_xreg = new_xreg.shape
                assert (
                    n_features_xreg == self.xreg.shape[1]
                ), "check number of series provided for 'new_xreg' (compare with self.xreg.shape[1])"
            except:
                n_obs_xreg = new_xreg.shape  # one series

            assert (
                n_obs_xreg == h
            ), "please provide values of regressors 'new_xreg' for the whole horizon 'h'"

            n_features_xreg = n_features_xreg * self.lags

            n_features_total = n_features + n_features_xreg

            inv_new_xreg = mo.rbind(self.xreg, new_xreg)[
                ::-1
            ]

        # Loop on horizon h

        for i in range(h):

            new_obs = ts.reformat_response(
                self.preds, self.lags
            )

            if (
                new_xreg is not None
            ):  # Additional regressors provided

                new_obs_xreg = ts.reformat_response(
                    inv_new_xreg, self.lags
                )

                new_X = mo.rbind(
                    np.union1d(
                        new_obs.reshape(1, n_features),
                        new_obs_xreg.reshape(
                            1, n_features_xreg
                        ),
                    ),
                    np.ones(n_features_total).reshape(
                        1, n_features_total
                    ),
                )

            else:

                new_X = mo.rbind(
                    new_obs.reshape(1, n_features),
                    np.ones(n_features).reshape(
                        1, n_features
                    ),
                )

            cooked_new_X = self.cook_test_set(
                new_X, **kwargs
            )

            predicted_cooked_new_X = self.obj.predict(
                cooked_new_X, **kwargs
            )

            if (
                self.return_std == False
            ):  # std. dev. is not returned

                preds = np.array(
                    [
                        (
                            self.y_means[j]
                            + predicted_cooked_new_X[0]
                        )
                        for j in range(self.n_series)
                    ]
                )

                self.preds = mo.rbind(preds, self.preds)

            else:  # std. dev. is returned

                preds = np.array(
                    [
                        (
                            self.y_means[j]
                            + predicted_cooked_new_X[0][0]
                        )
                        for j in range(self.n_series)
                    ]
                )

                self.preds = mo.rbind(preds, self.preds)

                self.preds_std[i] = predicted_cooked_new_X[
                    1
                ][0]

        # function's return

        self.preds = self.preds[0:h, :][::-1]

        if (
            self.return_std == False
        ):  # the std. dev. is returned

            return self.preds

        else:

            self.preds_std = self.preds_std[::-1].reshape(
                h, 1
            )

            # reshape
            self.preds_std = np.repeat(
                self.preds_std, self.n_series
            ).reshape(-1, self.n_series)

            return (
                self.preds,
                self.preds_std,
                self.preds - qt * self.preds_std,
                self.preds + qt * self.preds_std,
            )

    def score(
        self,
        X,
        training_index,
        testing_index,
        scoring=None,
        **kwargs
    ):
        """ Train on training_index, score on testing_index. """

        assert (
            bool(
                set(training_index).intersection(
                    set(testing_index)
                )
            )
            == False
        ), "Non-overlapping 'training_index' and 'testing_index' required"

        # Dimensions
        n, p = X.shape

        # Training and testing sets
        X_train = X[training_index, :]
        X_test = X[testing_index, :]

        # Horizon
        h = len(testing_index)
        assert (
            len(training_index) + h
        ) <= n, "Please check lengths of training and testing windows"

        # Fit and predict
        self.fit(X_train, **kwargs)
        preds = self.predict(
            h=h, return_std=False, **kwargs
        )

        if scoring is None:
            scoring = "neg_mean_squared_error"

        # check inputs
        assert scoring in (
            "explained_variance",
            "neg_mean_absolute_error",
            "neg_mean_squared_error",
            "neg_mean_squared_log_error",
            "neg_median_absolute_error",
            "r2",
        ), "'scoring' should be in ('explained_variance', 'neg_mean_absolute_error', \
                               'neg_mean_squared_error', 'neg_mean_squared_log_error', \
                               'neg_median_absolute_error', 'r2')"

        scoring_options = {
            "explained_variance": skm2.explained_variance_score,
            "neg_mean_absolute_error": skm2.mean_absolute_error,
            "neg_mean_squared_error": skm2.mean_squared_error,
            "neg_mean_squared_log_error": skm2.mean_squared_log_error,
            "neg_median_absolute_error": skm2.median_absolute_error,
            "r2": skm2.r2_score,
        }

        return tuple(
            [
                scoring_options[scoring](
                    X_test[:, i], preds[:, i], **kwargs
                )
                for i in range(p)
            ]
        )



from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="nnetsauce",
    version="0.1.0",
    url="https://github.com/thierrymoudiki/nnetsauce",
    packages=find_packages(),
    author="Thierry Moudiki",
    author_email="thierry.moudiki@gmail.com",
    description="Machine Learning using combinations of Neural Networks' layers",
    download_url="https://github.com/thierrymoudiki/nnetsauce",
    install_requires=["numpy >= 1.13.0", "scipy >= 0.19.0", "scikit-learn >= 0.18.0"],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
    ],
)



"""Custom model"""

# Authors: Thierry Moudiki
#
# License: BSD 3

from ..base import Base


class Custom(Base):
    """Custom model class derived from class Base
    
       Parameters
       ----------
       obj: object
           any object containing a method fit (obj.fit()) and a method predict 
           (obj.predict())
       n_hidden_features: int
           number of nodes in the hidden layer
       activation_name: str
           activation function: 'relu', 'tanh', 'sigmoid', 'prelu' or 'elu'
       a: float
           hyperparameter for 'prelu' or 'elu' activation function
       nodes_sim: str
           type of simulation for the nodes: 'sobol', 'hammersley', 'halton', 
           'uniform'
       bias: boolean
           indicates if the hidden layer contains a bias term (True) or not 
           (False)
       dropout: float
           regularization parameter; (random) percentage of nodes dropped out 
           of the training
       direct_link: boolean
           indicates if the original predictors are included (True) in model's 
           fitting or not (False)
       n_clusters: int
           number of clusters for 'kmeans' or 'gmm' clustering (could be 0: 
               no clustering)
       type_clust: str
           type of clustering method: currently k-means ('kmeans') or Gaussian 
           Mixture Model ('gmm')
       type_scaling: a tuple of 3 strings
           scaling methods for inputs, hidden layer, and clustering respectively
           (and when relevant). 
           Currently available: standardization ('std') or MinMax scaling ('minmax')
       col_sample: float
           percentage of covariates randomly chosen for training 
       row_sample: float
           percentage of rows chosen for training, by stratified bootstrapping    
       seed: int 
           reproducibility seed for nodes_sim=='uniform'
    """

    # construct the object -----

    def __init__(
        self,
        obj,
        n_hidden_features=5,
        activation_name="relu",
        a=0.01,
        nodes_sim="sobol",
        bias=True,
        dropout=0,
        direct_link=True,
        n_clusters=2,
        type_clust="kmeans",
        type_scaling=("std", "std", "std"),
        col_sample=1,
        row_sample=1,
        seed=123,
    ):

        super().__init__(
            n_hidden_features=n_hidden_features,
            activation_name=activation_name,
            a=a,
            nodes_sim=nodes_sim,
            bias=bias,
            dropout=dropout,
            direct_link=direct_link,
            n_clusters=n_clusters,
            type_clust=type_clust,
            type_scaling=type_scaling,
            col_sample=col_sample,
            row_sample=row_sample,
            seed=seed,
        )

        self.obj = obj



"""Random Vector Functional Link Network regression with 2 regularization parameters."""

# Authors: Thierry Moudiki
#
# License: BSD 3

import numpy as np
import sklearn.metrics as skm2
from ..base import Base
from ..utils import misc as mx
from ..utils import matrixops as mo
from ..utils import lmfuncs as lmf
from sklearn.base import RegressorMixin


class BayesianRVFL2Regressor(Base, RegressorMixin):
    """Bayesian RVFL model class derived from class Base, with parametrized covariance
    
       Parameters
       ----------
       n_hidden_features: int
           number of nodes in the hidden layer
       activation_name: str
           activation function: 'relu', 'tanh', 'sigmoid', 'prelu' or 'elu'
       a: float
           hyperparameter for 'prelu' or 'elu' activation function
       nodes_sim: str
           type of simulation for the nodes: 'sobol', 'hammersley', 'halton', 'uniform'
       bias: boolean
           indicates if the hidden layer contains a bias term (True) or not (False)
       dropout: float
           regularization parameter; (random) percentage of nodes dropped out 
           of the training
       direct_link: boolean
           indicates if the original predictors are included (True) in model's fitting or not (False)
       n_clusters: int
           number of clusters for 'kmeans' or 'gmm' clustering (could be 0: no clustering)
       type_clust: str
           type of clustering method: currently k-means ('kmeans') or Gaussian Mixture Model ('gmm')
       type_scaling: a tuple of 3 strings
           scaling methods for inputs, hidden layer, and clustering respectively
           (and when relevant). 
           Currently available: standardization ('std') or MinMax scaling ('minmax')
       col_sample: float
           percentage of covariates randomly chosen for training  
       row_sample: float
           percentage of rows chosen for training, by stratified bootstrapping    
       seed: int 
           reproducibility seed for nodes_sim=='uniform'
       s1: float
           std. dev. of init. regression parameters in Bayesian Ridge Regression
       s2: float
           std. dev. of augmented regression parameters in Bayesian Ridge Regression  
       sigma: float
           std. dev. of residuals in Bayesian Ridge Regression
       beta: array-like
           regression's fitted parameters 
       Sigma: array-like
           covariance of the distribution of fitted parameters
       GCV: float
       return_std: boolean
    """

    # construct the object -----

    def __init__(
        self,
        n_hidden_features=5,
        activation_name="relu",
        a=0.01,
        nodes_sim="sobol",
        bias=True,
        dropout=0,
        direct_link=True,
        n_clusters=0,
        type_clust="kmeans",
        type_scaling=("std", "std", "std"),
        col_sample=1,
        row_sample=1,
        seed=123,
        s1=0.1,
        s2=0.1,
        sigma=0.05,
        beta=None,
        Sigma=None,
        GCV=None,
        return_std=True,
    ):

        super().__init__(
            n_hidden_features=n_hidden_features,
            activation_name=activation_name,
            a=a,
            nodes_sim=nodes_sim,
            bias=bias,
            dropout=dropout,
            direct_link=direct_link,
            n_clusters=n_clusters,
            type_clust=type_clust,
            type_scaling=type_scaling,
            col_sample=col_sample,
            row_sample=row_sample,
            seed=seed,
        )

        self.s1 = s1
        self.s2 = s2
        self.sigma = sigma
        self.beta = beta
        self.Sigma = Sigma
        self.GCV = GCV
        self.return_std = return_std

    def fit(self, X, y, **kwargs):
        """Fit regularized RVFL to training data (X, y).
        
        Parameters
        ----------
        X: {array-like}, shape = [n_samples, n_features]
            Training vectors, where n_samples is the number 
            of samples and n_features is the number of features.
        
        y: array-like, shape = [n_samples]
               Target values.
    
        **kwargs: additional parameters to be passed to 
                  self.cook_training_set
               
        Returns
        -------
        self: object
        """

        centered_y, scaled_Z = self.cook_training_set(
            y=y, X=X, **kwargs
        )

        n, p = X.shape
        q = self.n_hidden_features

        if self.direct_link == True:

            r = p + self.n_clusters

            block11 = (self.s1 ** 2) * np.eye(r)
            block12 = np.zeros((r, q))
            block21 = np.zeros((q, r))
            block22 = (self.s2 ** 2) * np.eye(q)

            Sigma_prior = mo.rbind(
                mo.cbind(block11, block12),
                mo.cbind(block21, block22),
            )

        else:

            Sigma_prior = (self.s2 ** 2) * np.eye(q)

        fit_obj = lmf.beta_Sigma_hat_rvfl2(
            X=scaled_Z,
            y=centered_y,
            Sigma=Sigma_prior,
            sigma=self.sigma,
            fit_intercept=False,
            return_cov=self.return_std,
        )

        self.beta = fit_obj["beta_hat"]

        if self.return_std == True:
            self.Sigma = fit_obj["Sigma_hat"]

        self.GCV = fit_obj["GCV"]

        return self

    def predict(self, X, return_std=False, **kwargs):
        """Predict test data X.
        
        Parameters
        ----------
        X: {array-like}, shape = [n_samples, n_features]
            Training vectors, where n_samples is the number 
            of samples and n_features is the number of features.
        return_std: {boolean}, standard dev. is returned or not
        **kwargs: additional parameters to be passed to 
                  self.cook_test_set
               
        Returns
        -------
        model predictions: {array-like}
        """

        if (
            len(X.shape) == 1
        ):  # one observation in the test set only
            n_features = X.shape[0]
            new_X = mo.rbind(
                X.reshape(1, n_features),
                np.ones(n_features).reshape(1, n_features),
            )

        self.return_std = return_std

        if self.return_std == False:

            if len(X.shape) == 1:

                return (
                    self.y_mean
                    + np.dot(
                        self.cook_test_set(new_X, **kwargs),
                        self.beta,
                    )
                )[0]

            else:

                return self.y_mean + np.dot(
                    self.cook_test_set(X, **kwargs),
                    self.beta,
                )

        else:  # confidence interval required for preds?

            if len(X.shape) == 1:

                Z = self.cook_test_set(new_X, **kwargs)

                pred_obj = lmf.beta_Sigma_hat_rvfl2(
                    X_star=Z,
                    return_cov=self.return_std,
                    beta_hat_=self.beta,
                    Sigma_hat_=self.Sigma,
                )

                return (
                    self.y_mean + pred_obj["preds"][0],
                    pred_obj["preds_std"][0],
                )

            else:

                Z = self.cook_test_set(X, **kwargs)

                pred_obj = lmf.beta_Sigma_hat_rvfl2(
                    X_star=Z,
                    return_cov=self.return_std,
                    beta_hat_=self.beta,
                    Sigma_hat_=self.Sigma,
                )

                return (
                    self.y_mean + pred_obj["preds"],
                    pred_obj["preds_std"],
                )

    def score(self, X, y, scoring=None, **kwargs):
        """ Score the model on test set covariates X and response y. """

        preds = self.predict(X)

        if (
            self.return_std
        ):  # if there are std. devs in the predictions
            preds = preds[0]

        if mx.is_factor(y):  # classification

            if scoring is None:
                scoring = "accuracy"

            # check inputs
            assert scoring in (
                "accuracy",
                "average_precision",
                "brier_score_loss",
                "f1",
                "f1_micro",
                "f1_macro",
                "f1_weighted",
                "f1_samples",
                "neg_log_loss",
                "precision",
                "recall",
                "roc_auc",
            ), "'scoring' should be in ('accuracy', 'average_precision', \
                               'brier_score_loss', 'f1', 'f1_micro', \
                               'f1_macro', 'f1_weighted',  'f1_samples', \
                               'neg_log_loss', 'precision', 'recall', \
                               'roc_auc')"

            scoring_options = {
                "accuracy": skm2.accuracy_score,
                "average_precision": skm2.average_precision_score,
                "brier_score_loss": skm2.brier_score_loss,
                "f1": skm2.f1_score,
                "f1_micro": skm2.f1_score,
                "f1_macro": skm2.f1_score,
                "f1_weighted": skm2.f1_score,
                "f1_samples": skm2.f1_score,
                "neg_log_loss": skm2.log_loss,
                "precision": skm2.precision_score,
                "recall": skm2.recall_score,
                "roc_auc": skm2.roc_auc_score,
            }

        else:  # regression

            if scoring is None:
                scoring = "neg_mean_squared_error"

            # check inputs
            assert scoring in (
                "explained_variance",
                "neg_mean_absolute_error",
                "neg_mean_squared_error",
                "neg_mean_squared_log_error",
                "neg_median_absolute_error",
                "r2",
            ), "'scoring' should be in ('explained_variance', 'neg_mean_absolute_error', \
                               'neg_mean_squared_error', 'neg_mean_squared_log_error', \
                               'neg_median_absolute_error', 'r2')"

            scoring_options = {
                "explained_variance": skm2.explained_variance_score,
                "neg_mean_absolute_error": skm2.mean_absolute_error,
                "neg_mean_squared_error": skm2.mean_squared_error,
                "neg_mean_squared_log_error": skm2.mean_squared_log_error,
                "neg_median_absolute_error": skm2.median_absolute_error,
                "r2": skm2.r2_score,
            }

        return scoring_options[scoring](y, preds, **kwargs)
